C Copyright (c) 2012-2013 J. Scott Berg

C Permission is hereby granted, free of charge, to any person obtaining a copy
C of this software and associated documentation files (the "Software"), to deal
C in the Software without restriction, including without limitation the rights
C to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
C copies of the Software, and to permit persons to whom the Software is
C furnished to do so, subject to the following conditions:

C The above copyright notice and this permission notice shall be included in
C all copies or substantial portions of the Software.

C THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
C IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
C FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
C AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
C LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
C OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
C SOFTWARE.

C Random number generator based on 4-round AES-128 in counter mode

C Usage is as follows:

C ICOOL_IRAND must be called once before anything else

C SUBROUTINE ICOOL_SRAND(K)
C INTEGER K(4)
C   K (IN): can be initialized to anything; this is your seed
C Use K to initialize a state for usage in later calls to get random numbers

C FUNCTION ICOOL_DRAND()
C DOUBLE PRECISION ICOOL_DRAND
C Returns a double precision number uniformly distributed on [0,1)
C S must be initialized by ARS128_4_SRAND

      function gf2e8mul(a,b)

      implicit none
      integer a,b,gf2e8mul,m,i
      intrinsic btest,ieor,ishft
      data m/z'11b'/

      gf2e8mul = 0
      do 1 i=0,7
         if (btest(a,i)) gf2e8mul = ieor(gf2e8mul,ishft(b,i))
 1    continue
      do 2 i=6,0,-1
         if (btest(gf2e8mul,i+8)) gf2e8mul = ieor(gf2e8mul,ishft(m,i))
 2    continue

      end

      subroutine i8div(a,b,q,r)

      implicit none
      integer a,b,ia,ib,q,r
      intrinsic btest,ibset,ieor,ishft

      q = 0
      r = a
      
      ia = 0
 1    if (ishft(a,-ia).eq.0) goto 2
      ia = ia+1
      goto 1

 2    ib = 0
 3    if (ishft(b,-ib).eq.0) goto 4
      ib = ib+1
      goto 3

 4    if (ia.lt.ib) return
      if (btest(r,ia-1)) then
         q = ibset(q,ia-ib)
         r = ieor(r,ishft(b,ia-ib))
      endif
      ia = ia-1
      goto 4
      
      end

      function gf2e8inv(a)

      integer a,gf2e8inv,gf2e8mul,m,g0,g1,g2,c0a,c2a,q
      intrinsic ieor
      external gf2e8mul,i8div
      data m/z'11b'/

c     c_m*m + c_a*a = g_
      g0 = m
      g1 = a
      c0a = 0
      gf2e8inv = 1
 1    if (g1.eq.1) return
      call i8div(g0,g1,q,g2)
      c2a = ieor(c0a,gf2e8mul(q,gf2e8inv))
      c0a = gf2e8inv
      gf2e8inv = c2a
      g0 = g1
      g1 = g2
      goto 1

      end

      subroutine ars128_4_init

      implicit none
      integer c,i,gf2e8inv,gf2e8mul
      integer s1,s2,s3
      external gf2e8inv,gf2e8mul
      intrinsic ieor,ishftc
      data c/z'63'/

      integer rcon(4),rx(0:255,8)
      common /ars/rx,rcon
      save /ars/

      do 1 i=0,255
         if (i.eq.0) then
            s1 = c
         else
            s1 = gf2e8inv(i)
            s1 = ieor(c,ieor(s1,ieor(ishftc(s1,1,8),
     $           ieor(ishftc(s1,2,8),ieor(ishftc(s1,3,8),
     $           ishftc(s1,4,8))))))
         endif
         s1 = s1
         s2 = gf2e8mul(s1,2)
         s3 = gf2e8mul(s1,3)
         rx(i,1) = ieor(s2,ieor(ishft(s1,8),ieor(ishft(s1,16),
     $        ishft(s3,24))))
         rx(i,2) = ieor(s3,ieor(ishft(s2,8),ieor(ishft(s1,16),
     $        ishft(s1,24))))
         rx(i,3) = ieor(s1,ieor(ishft(s3,8),ieor(ishft(s2,16),
     $        ishft(s1,24))))
         rx(i,4) = ieor(s1,ieor(ishft(s1,8),ieor(ishft(s3,16),
     $        ishft(s2,24))))
         rx(i,5) = s1
         rx(i,6) = ishft(s1,8)
         rx(i,7) = ishft(s1,16)
         rx(i,8) = ishft(s1,24)
 1    continue
      rcon(1) = 1
      do 3 i=2,4
         rcon(i) = gf2e8mul(rcon(i-1),2)
 3    continue

      end

      subroutine ars128_4_ke(k,w)

      implicit none
      integer nb,nk,nr
      parameter (nb=4,nk=4,nr=4)
      integer k(nk),w(nb*(nr+1))
      intrinsic ibits,ieor,ishftc,mvbits
      integer i

      integer rcon(4),rx(0:255,8)
      common /ars/rx,rcon
      save /ars/

      do 1 i=1,nk
         w(i) = k(i)
 1    continue
      do 2 i=nk+1,nb*(nr+1)
         if (mod(i,nk).eq.1) then
            w(i) = ieor(rcon(i/nk),ieor(rx(ibits(w(i-1),8,8),5),ieor(
     $           rx(ibits(w(i-1),16,8),6),ieor(
     $           rx(ibits(w(i-1),24,8),7),rx(ibits(w(i-1),0,8),8)))))
         else
            w(i) = w(i-1)
         endif
         w(i) = ieor(w(i),w(i-nk))
 2    continue

      end

      subroutine ars128_4_cipher(in,out,w)

      implicit none
      integer nb,nr
      parameter (nb=4,nr=4)
      integer in(nb),out(nb),w(nb*(nr+1)),t(nb,2)
      integer i,j,k
      intrinsic ibits,ieor,mvbits
      
      integer rcon(4),rx(0:255,8)
      common /ars/rx,rcon
      save /ars/

      do 10 j=1,nb
         t(j,1) = ieor(in(j),w(j))
 10   continue
      k = 1
      do 2 i=1,nr-1
         t(1,3-k) = ieor(ieor(ieor(ieor(rx(ibits(t(1,k),0,8),1),
     $        rx(ibits(t(2,k),8,8),2)),rx(ibits(t(3,k),16,8),3)),
     $        rx(ibits(t(4,k),24,8),4)),w(j))
         j = j+1
         t(2,3-k) = ieor(ieor(ieor(ieor(rx(ibits(t(2,k),0,8),1),
     $        rx(ibits(t(3,k),8,8),2)),rx(ibits(t(4,k),16,8),3)),
     $        rx(ibits(t(1,k),24,8),4)),w(j))
         j = j+1
         t(3,3-k) = ieor(ieor(ieor(ieor(rx(ibits(t(3,k),0,8),1),
     $        rx(ibits(t(4,k),8,8),2)),rx(ibits(t(1,k),16,8),3)),
     $        rx(ibits(t(2,k),24,8),4)),w(j))
         j = j+1
         t(4,3-k) = ieor(ieor(ieor(ieor(rx(ibits(t(4,k),0,8),1),
     $        rx(ibits(t(1,k),8,8),2)),rx(ibits(t(2,k),16,8),3)),
     $        rx(ibits(t(3,k),24,8),4)),w(j))
         j = j+1
         k = 3-k
 2    continue
      out(1) = ieor(ieor(ieor(ieor(rx(ibits(t(1,k),0,8),5),
     $     rx(ibits(t(2,k),8,8),6)),rx(ibits(t(3,k),16,8),7)),
     $     rx(ibits(t(4,k),24,8),8)),w(j))
      j = j+1
      out(2) = ieor(ieor(ieor(ieor(rx(ibits(t(2,k),0,8),5),
     $     rx(ibits(t(3,k),8,8),6)),rx(ibits(t(4,k),16,8),7)),
     $     rx(ibits(t(1,k),24,8),8)),w(j))
      j = j+1
      out(3) = ieor(ieor(ieor(ieor(rx(ibits(t(3,k),0,8),5),
     $     rx(ibits(t(4,k),8,8),6)),rx(ibits(t(1,k),16,8),7)),
     $     rx(ibits(t(2,k),24,8),8)),w(j))
      j = j+1
      out(4) = ieor(ieor(ieor(ieor(rx(ibits(t(4,k),0,8),5),
     $     rx(ibits(t(1,k),8,8),6)),rx(ibits(t(2,k),16,8),7)),
     $     rx(ibits(t(3,k),24,8),8)),w(j))

      end

      subroutine icool_srand(k)

      implicit none
      integer nb,nr
      parameter (nb=4,nr=4)
      integer k(nb),s(nb*(nr+3)+1)
      integer i
      external ars128_4_cipher,ars128_4_ke,ars128_4_init
      common /icool_ars/s
      save /icool_ars/

      call ars128_4_init
      s(1) = 2
      do 1 i=2,1+nb
         s(i) = 0
 1    continue
      call ars128_4_ke(k,s(2+2*nb))
      call ars128_4_cipher(s(2),s(2+nb),s(2+2*nb))

      end

      function icool_randget()

      implicit none
      integer nb,nr
      parameter (nb=4,nr=4)

      integer icool_randget

      integer j
      integer t

      integer s(nb*(nr+3)+1)

      intrinsic btest,ibits,mvbits
      external ars128_4_cipher

      common /icool_ars/s
      save /icool_ars/

      icool_randget = s(nb+s(1))
      s(1) = s(1)+1
      if (s(1).eq.6) then
         j = 2
 2       t = ibits(s(j),0,16)+1
         call mvbits(t,0,16,s(j),0)
         if (btest(t,16)) then
            t = ibits(s(j),16,16)+1
            call mvbits(t,0,16,s(j),16)
            if (btest(t,16).and.j.lt.5) then
               j = j+1
               goto 2
            endif
         endif
         call ars128_4_cipher(s(2),s(2+nb),s(2+2*nb))
         s(1) = 2
      endif
      
      end

      function icool_drand()

      implicit none

      integer icool_randget,t
      double precision icool_drand
      external icool_randget
      intrinsic ibits

      t = icool_randget()
      icool_drand = ibits(t,0,16)*2d0**(-16)
     $     +ibits(t,16,16)*2d0**(-32)

      end
